#! /usr/bin/env python3
# Script for preparing OpenSSL for building on Windows.
# Uses Perl to create nmake makefiles and otherwise prepare the way
# for building on 32 or 64 bit platforms.

# Script originally authored by Mark Hammond.
# Major revisions by:
#   Martin v. Löwis
#   Christian Heimes
#   Zachary Ware

# THEORETICALLY, you can:
# * Unpack the latest OpenSSL release where $(opensslDir) in
#   PCbuild\pyproject.props expects it to be.
# * Install ActivePerl and ensure it is somewhere on your path.
# * Run this script with the OpenSSL source dir as the only argument.
#
# it should configure OpenSSL such that it is ready to be built by
# ssl.vcxproj on 32 or 64 bit platforms.

from __future__ import print_function

import os
import re
import sys
import subprocess
from shutil import copy

# Find all "foo.exe" files on the PATH.
def find_all_on_path(filename, extras=None):
    entries = os.environ["PATH"].split(os.pathsep)
    ret = []
    for p in entries:
        fname = os.path.abspath(os.path.join(p, filename))
        if os.path.isfile(fname) and fname not in ret:
            ret.append(fname)
    if extras:
        for p in extras:
            fname = os.path.abspath(os.path.join(p, filename))
            if os.path.isfile(fname) and fname not in ret:
                ret.append(fname)
    return ret


# Find a suitable Perl installation for OpenSSL.
# cygwin perl does *not* work.  ActivePerl does.
# Being a Perl dummy, the simplest way I can check is if the "Win32" package
# is available.
def find_working_perl(perls):
    for perl in perls:
        try:
            subprocess.check_output([perl, "-e", "use Win32;"])
        except subprocess.CalledProcessError:
            continue
        else:
            return perl

    if perls:
        print("The following perl interpreters were found:")
        for p in perls:
            print(" ", p)
        print(" None of these versions appear suitable for building OpenSSL")
    else:
        print("NO perl interpreters were found on this machine at all!")
    print(" Please install ActivePerl and ensure it appears on your path")


def copy_includes(makefile, suffix):
    dir = 'inc'+suffix+'\\openssl'
    try:
        os.makedirs(dir)
    except OSError:
        pass
    copy_if_different = r'$(PERL) $(SRC_D)\util\copy-if-different.pl'
    with open(makefile) as fin:
        for line in fin:
            if copy_if_different in line:
                perl, script, src, dest = line.split()
                if not '$(INCO_D)' in dest:
                    continue
                # We're in the root of the source tree
                src = src.replace('$(SRC_D)', '.').strip('"')
                dest = dest.strip('"').replace('$(INCO_D)', dir)
                print('copying', src, 'to', dest)
                copy(src, dest)


def run_configure(configure, do_script):
    print("perl Configure "+configure+" no-idea no-mdc2")
    os.system("perl Configure "+configure+" no-idea no-mdc2")
    print(do_script)
    os.system(do_script)

def fix_uplink():
    # uplink.c tries to find the OPENSSL_Applink function exported from the current
    # executable. However, we export it from _ssl[_d].pyd instead. So we update the
    # module name here before building.
    with open('ms\\uplink.c', 'r', encoding='utf-8') as f1:
        code = list(f1)
    os.replace('ms\\uplink.c', 'ms\\uplink.c.orig')
    already_patched = False
    with open('ms\\uplink.c', 'w', encoding='utf-8') as f2:
        for line in code:
            if not already_patched:
                if re.search('MODIFIED FOR CPYTHON _ssl MODULE', line):
                    already_patched = True
                elif re.match(r'^\s+if\s*\(\(h\s*=\s*GetModuleHandle[AW]?\(NULL\)\)\s*==\s*NULL\)', line):
                    f2.write("/* MODIFIED FOR CPYTHON _ssl MODULE */\n")
                    f2.write('if ((h = GetModuleHandleW(L"_ssl.pyd")) == NULL) if ((h = GetModuleHandleW(L"_ssl_d.pyd")) == NULL)\n')
                    already_patched = True
            f2.write(line)
    if not already_patched:
        print("WARN: failed to patch ms\\uplink.c")

def prep(arch):
    makefile_template = "ms\\ntdll{}.mak"
    generated_makefile = makefile_template.format('')
    if arch == "x86":
        configure = "VC-WIN32"
        do_script = "ms\\do_nasm"
        suffix = "32"
    elif arch == "amd64":
        configure = "VC-WIN64A"
        do_script = "ms\\do_win64a"
        suffix = "64"
    else:
        raise ValueError('Unrecognized platform: %s' % arch)

    print("Creating the makefiles...")
    sys.stdout.flush()
    # run configure, copy includes, patch files
    run_configure(configure, do_script)
    makefile = makefile_template.format(suffix)
    try:
        os.unlink(makefile)
    except FileNotFoundError:
        pass
    os.rename(generated_makefile, makefile)
    copy_includes(makefile, suffix)

    print('patching ms\\uplink.c...')
    fix_uplink()

def main():
    if len(sys.argv) == 1:
        print("Not enough arguments: directory containing OpenSSL",
              "sources must be supplied")
        sys.exit(1)

    if len(sys.argv) == 3 and sys.argv[2] not in ('x86', 'amd64'):
        print("Second argument must be x86 or amd64")
        sys.exit(1)

    if len(sys.argv) > 3:
        print("Too many arguments supplied, all we need is the directory",
              "containing OpenSSL sources and optionally the architecture")
        sys.exit(1)

    ssl_dir = sys.argv[1]
    arch = sys.argv[2] if len(sys.argv) >= 3 else None

    if not os.path.isdir(ssl_dir):
        print(ssl_dir, "is not an existing directory!")
        sys.exit(1)

    # perl should be on the path, but we also look in "\perl" and "c:\\perl"
    # as "well known" locations
    perls = find_all_on_path("perl.exe", [r"\perl\bin",
                                          r"C:\perl\bin",
                                          r"\perl64\bin",
                                          r"C:\perl64\bin",
                                         ])
    perl = find_working_perl(perls)
    if perl:
        print("Found a working perl at '%s'" % (perl,))
    else:
        sys.exit(1)
    if not find_all_on_path('nmake.exe'):
        print('Could not find nmake.exe, try running env.bat')
        sys.exit(1)
    if not find_all_on_path('nasm.exe'):
        print('Could not find nasm.exe, please add to PATH')
        sys.exit(1)
    sys.stdout.flush()

    # Put our working Perl at the front of our path
    os.environ["PATH"] = os.path.dirname(perl) + \
                                os.pathsep + \
                                os.environ["PATH"]

    old_cwd = os.getcwd()
    try:
        os.chdir(ssl_dir)
        if arch:
            prep(arch)
        else:
            for arch in ['amd64', 'x86']:
                prep(arch)
    finally:
        os.chdir(old_cwd)

if __name__=='__main__':
    main()



#! /usr/bin/env python3

"""
This script should be called *manually* when we want to upgrade SSLError
`library` and `reason` mnemonics to a more recent OpenSSL version.

It takes two arguments:
- the path to the OpenSSL source tree (e.g. git checkout)
- the path to the C file to be generated
  (probably Modules/_ssl_data.h)
"""

import datetime
import glob
import os
import re
import sys
import _ssl


def parse_error_codes(h_file, prefix, libcode):
    pat = re.compile(r"#\s*define\W+(%s([\w]+))\W+(\d+)\b" % re.escape(prefix))
    codes = []
    with open(h_file, "r", encoding="latin1") as f:
        for line in f:
            match = pat.search(line)
            if match:
                code, name, num = match.groups()
                num = int(num)
                # e.g. ("SSL_R_BAD_DATA", ("ERR_LIB_SSL", "BAD_DATA", 390))
                codes.append((code, (libcode, name, num)))
    assert codes, f"no codes found in {h_file}"
    return codes

if __name__ == "__main__":
    openssl_inc = sys.argv[1]
    outfile = sys.argv[2]
    use_stdout = outfile == '-'
    f = sys.stdout if use_stdout else open(outfile, "w")
    # mnemonic -> (library code, error prefix, header file)
    error_libraries = {}
    for error_header in glob.glob(os.path.join(glob.escape(openssl_inc), 'include/openssl/*err.h')):
        base = os.path.basename(error_header)
        if base in ('buffererr.h', 'objectserr.h', 'storeerr.h'):
            # Deprecated in 3.0.
            continue
        mnemonic = base[:-5].upper()
        if mnemonic == "":
            # err.h
            lib_codes = {
                code: num
                for (code, (_, _, num)) in parse_error_codes(error_header, 'ERR_LIB_', None)
            }
        else:
            error_libraries[mnemonic] = (f'ERR_LIB_{mnemonic}', f'{mnemonic}_R_', error_header)

    # Read codes from libraries
    new_codes = []
    for libcode, prefix, h_file in sorted(error_libraries.values()):
        new_codes += parse_error_codes(h_file, prefix, libcode)
    new_code_nums = set((libcode, num)
                        for (code, (libcode, name, num)) in new_codes)

    # Merge with existing codes (in case some old codes disappeared).
    codes = {}
    for errname, (libnum, errnum) in _ssl.err_names_to_codes.items():
        lib = error_libraries[_ssl.lib_codes_to_names[libnum]]
        libcode = lib[0]              # e.g. ERR_LIB_PEM
        errcode = lib[1] + errname    # e.g. SSL_R_BAD_SSL_SESSION_ID_LENGTH
        # Only keep it if the numeric codes weren't reused
        if (libcode, errnum) not in new_code_nums:
            codes[errcode] = libcode, errname, errnum
    codes.update(dict(new_codes))

    def w(l):
        f.write(l + "\n")
    w("/* File generated by Tools/ssl/make_ssl_data.py */")
    w("/* Generated on %s */" % datetime.datetime.now().isoformat())
    w("")

    w("static struct py_ssl_library_code library_codes[] = {")
    for mnemo, (libcode, _, _) in sorted(error_libraries.items()):
        w(f'#ifdef {libcode}')
        w('    {"%s", %s},' % (mnemo, libcode))
        w('#endif')
    w('    { NULL }')
    w('};')
    w("")

    w("static struct py_ssl_error_code error_codes[] = {")
    for errcode, (libcode, name, num) in sorted(codes.items()):
        w('  #ifdef %s' % (errcode))
        w('    {"%s", %s, %s},' % (name, libcode, errcode))
        w('  #else')
        w('    {"%s", %s, %d},' % (name, lib_codes[libcode], num))
        w('  #endif')
    w('    { NULL }')
    w('};')
    if not use_stdout:
        f.close()



#! /usr/bin/env python3

# ptags
#
# Create a tags file for Python programs, usable with vi.
# Tagged are:
# - functions (even inside other defs or classes)
# - classes
# - filenames
# Warns about files it cannot open.
# No warnings about duplicate tags.

import sys, re, os

tags = []    # Modified global variable!

def main():
    args = sys.argv[1:]
    for filename in args:
        treat_file(filename)
    if tags:
        with open('tags', 'w') as fp:
            tags.sort()
            for s in tags: fp.write(s)


expr = r'^[ \t]*(def|class)[ \t]+([a-zA-Z0-9_]+)[ \t]*[:\(]'
matcher = re.compile(expr)

def treat_file(filename):
    try:
        fp = open(filename, 'r')
    except:
        sys.stderr.write('Cannot open %s\n' % filename)
        return
    with fp:
        base = os.path.basename(filename)
        if base[-3:] == '.py':
            base = base[:-3]
        s = base + '\t' + filename + '\t' + '1\n'
        tags.append(s)
        while 1:
            line = fp.readline()
            if not line:
                break
            m = matcher.match(line)
            if m:
                content = m.group(0)
                name = m.group(2)
                s = name + '\t' + filename + '\t/^' + content + '/\n'
                tags.append(s)

if __name__ == '__main__':
    main()



import functools
import hashlib
import unittest

try:
    import _hashlib
except ImportError:
    _hashlib = None


def requires_hashdigest(digestname, openssl=None, usedforsecurity=True):
    """Decorator raising SkipTest if a hashing algorithm is not available

    The hashing algorithm could be missing or blocked by a strict crypto
    policy.

    If 'openssl' is True, then the decorator checks that OpenSSL provides
    the algorithm. Otherwise the check falls back to built-in
    implementations. The usedforsecurity flag is passed to the constructor.

    ValueError: [digital envelope routines: EVP_DigestInit_ex] disabled for FIPS
    ValueError: unsupported hash type md4
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                if openssl and _hashlib is not None:
                    _hashlib.new(digestname, usedforsecurity=usedforsecurity)
                else:
                    hashlib.new(digestname, usedforsecurity=usedforsecurity)
            except ValueError:
                raise unittest.SkipTest(
                    f"hash digest '{digestname}' is not available."
                )
            return func(*args, **kwargs)
        return wrapper
    return decorator



import argparse
import re
import sys

from c_analyzer.common import show
from c_analyzer.common.info import UNKNOWN

from . import SOURCE_DIRS
from .find import supported_vars
from .known import (
    from_file as known_from_file,
    DATA_FILE as KNOWN_FILE,
    )
from .supported import IGNORED_FILE


def _check_results(unknown, knownvars, used):
    def _match_unused_global(variable):
        found = []
        for varid in knownvars:
            if varid in used:
                continue
            if varid.funcname is not None:
                continue
            if varid.name != variable.name:
                continue
            if variable.filename and variable.filename != UNKNOWN:
                if variable.filename == varid.filename:
                    found.append(varid)
            else:
                found.append(varid)
        return found

    badknown = set()
    for variable in sorted(unknown):
        msg = None
        if variable.funcname != UNKNOWN:
            msg = f'could not find global symbol {variable.id}'
        elif m := _match_unused_global(variable):
            assert isinstance(m, list)
            badknown.update(m)
        elif variable.name in ('completed', 'id'):  # XXX Figure out where these variables are.
            unknown.remove(variable)
        else:
            msg = f'could not find local symbol {variable.id}'
        if msg:
            #raise Exception(msg)
            print(msg)
    if badknown:
        print('---')
        print(f'{len(badknown)} globals in known.tsv, but may actually be local:')
        for varid in sorted(badknown):
            print(f'{varid.filename:30} {varid.name}')
    unused = sorted(varid
                    for varid in set(knownvars) - used
                    if varid.name != 'id')  # XXX Figure out where these variables are.
    if unused:
        print('---')
        print(f'did not use {len(unused)} known vars:')
        for varid in unused:
            print(f'{varid.filename:30} {varid.funcname or "-":20} {varid.name}')
        raise Exception('not all known symbols used')
    if unknown:
        print('---')
        raise Exception('could not find all symbols')


# XXX Move this check to its own command.
def cmd_check_cache(cmd, *,
                    known=KNOWN_FILE,
                    ignored=IGNORED_FILE,
                    _known_from_file=known_from_file,
                    _find=supported_vars,
                    ):
    known = _known_from_file(known)

    used = set()
    unknown = set()
    for var, supported in _find(known=known, ignored=ignored):
        if supported is None:
            unknown.add(var)
            continue
        used.add(var.id)
    _check_results(unknown, known['variables'], used)


def cmd_check(cmd, *,
              known=KNOWN_FILE,
              ignored=IGNORED_FILE,
              _find=supported_vars,
              _show=show.basic,
              _print=print,
              ):
    """
    Fail if there are unsupported globals variables.

    In the failure case, the list of unsupported variables
    will be printed out.
    """
    unsupported = []
    for var, supported in _find(known=known, ignored=ignored):
        if not supported:
            unsupported.append(var)

    if not unsupported:
        #_print('okay')
        return

    _print('ERROR: found unsupported global variables')
    _print()
    _show(sorted(unsupported))
    _print(f' ({len(unsupported)} total)')
    sys.exit(1)


def cmd_show(cmd, *,
             known=KNOWN_FILE,
             ignored=IGNORED_FILE,
             skip_objects=False,
              _find=supported_vars,
             _show=show.basic,
             _print=print,
             ):
    """
    Print out the list of found global variables.

    The variables will be distinguished as "supported" or "unsupported".
    """
    allsupported = []
    allunsupported = []
    for found, supported in _find(known=known,
                                  ignored=ignored,
                                  skip_objects=skip_objects,
                                  ):
        if supported is None:
            continue
        (allsupported if supported else allunsupported
         ).append(found)

    _print('supported:')
    _print('----------')
    _show(sorted(allsupported))
    _print(f' ({len(allsupported)} total)')
    _print()
    _print('unsupported:')
    _print('------------')
    _show(sorted(allunsupported))
    _print(f' ({len(allunsupported)} total)')


#############################
# the script

COMMANDS = {
        'check': cmd_check,
        'show': cmd_show,
        }

PROG = sys.argv[0]
PROG = 'c-globals.py'


def parse_args(prog=PROG, argv=sys.argv[1:], *, _fail=None):
    common = argparse.ArgumentParser(add_help=False)
    common.add_argument('--ignored', metavar='FILE',
                        default=IGNORED_FILE,
                        help='path to file that lists ignored vars')
    common.add_argument('--known', metavar='FILE',
                        default=KNOWN_FILE,
                        help='path to file that lists known types')
    #common.add_argument('dirs', metavar='DIR', nargs='*',
    #                    default=SOURCE_DIRS,
    #                    help='a directory to check')

    parser = argparse.ArgumentParser(
            prog=prog,
            )
    subs = parser.add_subparsers(dest='cmd')

    check = subs.add_parser('check', parents=[common])

    show = subs.add_parser('show', parents=[common])
    show.add_argument('--skip-objects', action='store_true')

    if _fail is None:
        def _fail(msg):
            parser.error(msg)

    # Now parse the args.
    args = parser.parse_args(argv)
    ns = vars(args)

    cmd = ns.pop('cmd')
    if not cmd:
        _fail('missing command')

    return cmd, ns


def main(cmd, cmdkwargs=None, *, _COMMANDS=COMMANDS):
    try:
        cmdfunc = _COMMANDS[cmd]
    except KeyError:
        raise ValueError(
            f'unsupported cmd {cmd!r}' if cmd else 'missing cmd')

    cmdfunc(cmd, **cmdkwargs or {})


if __name__ == '__main__':
    cmd, cmdkwargs = parse_args()
    main(cmd, cmdkwargs)



import unittest

from .. import tool_imports_for_tests
with tool_imports_for_tests():
    pass


class SelfCheckTests(unittest.TestCase):

    @unittest.expectedFailure
    def test_known(self):
        # Make sure known macros & vartypes aren't hiding unknown local types.
        # XXX finish!
        raise NotImplementedError

    @unittest.expectedFailure
    def test_compare_nm_results(self):
        # Make sure the "show" results match the statics found by "nm" command.
        # XXX Skip if "nm" is not available.
        # XXX finish!
        raise NotImplementedError


class DummySourceTests(unittest.TestCase):

    @unittest.expectedFailure
    def test_check(self):
        # XXX finish!
        raise NotImplementedError

    @unittest.expectedFailure
    def test_show(self):
        # XXX finish!
        raise NotImplementedError



import os.path
import shutil

from c_analyzer.common import util, info

from .info import Symbol


# XXX need tests:
# * iter_symbols

NM_KINDS = {
        'b': Symbol.KIND.VARIABLE,  # uninitialized
        'd': Symbol.KIND.VARIABLE,  # initialized
        #'g': Symbol.KIND.VARIABLE,  # uninitialized
        #'s': Symbol.KIND.VARIABLE,  # initialized
        't': Symbol.KIND.FUNCTION,
        }

SPECIAL_SYMBOLS = {
        # binary format (e.g. ELF)
        '__bss_start',
        '__data_start',
        '__dso_handle',
        '_DYNAMIC',
        '_edata',
        '_end',
        '__environ@@GLIBC_2.2.5',
        '_GLOBAL_OFFSET_TABLE_',
        '__JCR_END__',
        '__JCR_LIST__',
        '__TMC_END__',
        }


def _is_special_symbol(name):
    if name in SPECIAL_SYMBOLS:
        return True
    if '@@GLIBC' in name:
        return True
    return False


def iter_symbols(binfile, *,
                 nm=None,
                 handle_id=None,
                 _which=shutil.which,
                 _run=util.run_cmd,
                 ):
    """Yield a Symbol for each relevant entry reported by the "nm" command."""
    if nm is None:
        nm = _which('nm')
        if not nm:
            raise NotImplementedError
    if handle_id is None:
        handle_id = info.ID

    argv = [nm,
            '--line-numbers',
            binfile,
            ]
    try:
        output = _run(argv)
    except Exception:
        if nm is None:
            # XXX Use dumpbin.exe /SYMBOLS on Windows.
            raise NotImplementedError
        raise
    for line in output.splitlines():
        (name, kind, external, filename, funcname,
         ) = _parse_nm_line(line)
        if kind != Symbol.KIND.VARIABLE:
            continue
        elif _is_special_symbol(name):
            continue
        yield Symbol(
                id=handle_id(filename, funcname, name),
                kind=kind,
                external=external,
                )


def _parse_nm_line(line):
    _origline = line
    _, _, line = line.partition(' ')  # strip off the address
    line = line.strip()

    kind, _, line = line.partition(' ')
    line = line.strip()
    external = kind.isupper()
    kind = NM_KINDS.get(kind.lower(), Symbol.KIND.OTHER)

    name, _, filename = line.partition('\t')
    name = name.strip()
    if filename:
        filename = os.path.relpath(filename.partition(':')[0])
    else:
        filename = info.UNKNOWN

    name, islocal = _parse_nm_name(name, kind)
    funcname = info.UNKNOWN if islocal else None
    return name, kind, external, filename, funcname


def _parse_nm_name(name, kind):
    if kind != Symbol.KIND.VARIABLE:
        return name, None
    if _is_special_symbol(name):
        return name, None

    actual, sep, digits = name.partition('.')
    if not sep:
        return name, False

    if not digits.isdigit():
        raise Exception(f'got bogus name {name}')
    return actual, True



import os.path
import unittest

from .. import tool_imports_for_tests
with tool_imports_for_tests():
    from c_analyzer.common.files import (
            iter_files, _walk_tree, glob_tree,
            )


def fixpath(filename):
    return filename.replace('/', os.path.sep)


class IterFilesTests(unittest.TestCase):

    maxDiff = None

    _return_walk = None

    @property
    def calls(self):
        try:
            return self._calls
        except AttributeError:
            self._calls = []
            return self._calls

    def set_files(self, *filesperroot):
        roots = []
        result = []
        for root, files in filesperroot:
            root = fixpath(root)
            roots.append(root)
            result.append([os.path.join(root, fixpath(f))
                           for f in files])
        self._return_walk = result
        return roots

    def _walk(self, root, *, suffix=None, walk=None):
        self.calls.append(('_walk', (root, suffix, walk)))
        return iter(self._return_walk.pop(0))

    def _glob(self, root, *, suffix=None):
        self.calls.append(('_glob', (root, suffix)))
        return iter(self._return_walk.pop(0))

    def test_typical(self):
        dirnames = self.set_files(
            ('spam', ['file1.c', 'file2.c']),
            ('eggs', ['ham/file3.h']),
            )
        suffixes = ('.c', '.h')

        files = list(iter_files(dirnames, suffixes,
                                _glob=self._glob,
                                _walk=self._walk))

        self.assertEqual(files, [
            fixpath('spam/file1.c'),
            fixpath('spam/file2.c'),
            fixpath('eggs/ham/file3.h'),
            ])
        self.assertEqual(self.calls, [
            ('_walk', ('spam', None, _walk_tree)),
            ('_walk', ('eggs', None, _walk_tree)),
            ])

    def test_single_root(self):
        self._return_walk = [
                [fixpath('spam/file1.c'), fixpath('spam/file2.c')],
                ]

        files = list(iter_files('spam', '.c',
                                _glob=self._glob,
                                _walk=self._walk))

        self.assertEqual(files, [
            fixpath('spam/file1.c'),
            fixpath('spam/file2.c'),
            ])
        self.assertEqual(self.calls, [
            ('_walk', ('spam', '.c', _walk_tree)),
            ])

    def test_one_root(self):
        self._return_walk = [
                [fixpath('spam/file1.c'), fixpath('spam/file2.c')],
                ]

        files = list(iter_files(['spam'], '.c',
                                _glob=self._glob,
                                _walk=self._walk))

        self.assertEqual(files, [
            fixpath('spam/file1.c'),
            fixpath('spam/file2.c'),
            ])
        self.assertEqual(self.calls, [
            ('_walk', ('spam', '.c', _walk_tree)),
            ])

    def test_multiple_roots(self):
        dirnames = self.set_files(
            ('spam', ['file1.c', 'file2.c']),
            ('eggs', ['ham/file3.c']),
            )

        files = list(iter_files(dirnames, '.c',
                                _glob=self._glob,
                                _walk=self._walk))

        self.assertEqual(files, [
            fixpath('spam/file1.c'),
            fixpath('spam/file2.c'),
            fixpath('eggs/ham/file3.c'),
            ])
        self.assertEqual(self.calls, [
            ('_walk', ('spam', '.c', _walk_tree)),
            ('_walk', ('eggs', '.c', _walk_tree)),
            ])

    def test_no_roots(self):
        files = list(iter_files([], '.c',
                                _glob=self._glob,
                                _walk=self._walk))

        self.assertEqual(files, [])
        self.assertEqual(self.calls, [])

    def test_single_suffix(self):
        self._return_walk = [
                [fixpath('spam/file1.c'),
                 fixpath('spam/eggs/file3.c'),
                 ],
                ]

        files = list(iter_files('spam', '.c',
                                _glob=self._glob,
                                _walk=self._walk))

        self.assertEqual(files, [
            fixpath('spam/file1.c'),
            fixpath('spam/eggs/file3.c'),
            ])
        self.assertEqual(self.calls, [
            ('_walk', ('spam', '.c', _walk_tree)),
            ])

    def test_one_suffix(self):
        self._return_walk = [
                [fixpath('spam/file1.c'),
                 fixpath('spam/file1.h'),
                 fixpath('spam/file1.o'),
                 fixpath('spam/eggs/file3.c'),
                 ],
                ]

        files = list(iter_files('spam', ['.c'],
                                _glob=self._glob,
                                _walk=self._walk))

        self.assertEqual(files, [
            fixpath('spam/file1.c'),
            fixpath('spam/eggs/file3.c'),
            ])
        self.assertEqual(self.calls, [
            ('_walk', ('spam', None, _walk_tree)),
            ])

    def test_multiple_suffixes(self):
        self._return_walk = [
                [fixpath('spam/file1.c'),
                 fixpath('spam/file1.h'),
                 fixpath('spam/file1.o'),
                 fixpath('spam/eggs/file3.c'),
                 ],
                ]

        files = list(iter_files('spam', ('.c', '.h'),
                                _glob=self._glob,
                                _walk=self._walk))

        self.assertEqual(files, [
            fixpath('spam/file1.c'),
            fixpath('spam/file1.h'),
            fixpath('spam/eggs/file3.c'),
            ])
        self.assertEqual(self.calls, [
            ('_walk', ('spam', None, _walk_tree)),
            ])

    def test_no_suffix(self):
        expected = [fixpath('spam/file1.c'),
                    fixpath('spam/file1.h'),
                    fixpath('spam/file1.o'),
                    fixpath('spam/eggs/file3.c'),
                    ]
        for suffix in (None, '', ()):
            with self.subTest(suffix):
                self.calls.clear()
                self._return_walk = [list(expected)]

                files = list(iter_files('spam', suffix,
                                        _glob=self._glob,
                                        _walk=self._walk))

                self.assertEqual(files, expected)
                self.assertEqual(self.calls, [
                    ('_walk', ('spam', suffix, _walk_tree)),
                    ])

    def test_relparent(self):
        dirnames = self.set_files(
            ('/x/y/z/spam', ['file1.c', 'file2.c']),
            ('/x/y/z/eggs', ['ham/file3.c']),
            )

        files = list(iter_files(dirnames, '.c', fixpath('/x/y'),
                                _glob=self._glob,
                                _walk=self._walk))

        self.assertEqual(files, [
            fixpath('z/spam/file1.c'),
            fixpath('z/spam/file2.c'),
            fixpath('z/eggs/ham/file3.c'),
            ])
        self.assertEqual(self.calls, [
            ('_walk', (fixpath('/x/y/z/spam'), '.c', _walk_tree)),
            ('_walk', (fixpath('/x/y/z/eggs'), '.c', _walk_tree)),
            ])

    def test_glob(self):
        dirnames = self.set_files(
            ('spam', ['file1.c', 'file2.c']),
            ('eggs', ['ham/file3.c']),
            )

        files = list(iter_files(dirnames, '.c',
                                get_files=glob_tree,
                                _walk=self._walk,
                                _glob=self._glob))

        self.assertEqual(files, [
            fixpath('spam/file1.c'),
            fixpath('spam/file2.c'),
            fixpath('eggs/ham/file3.c'),
            ])
        self.assertEqual(self.calls, [
            ('_glob', ('spam', '.c')),
            ('_glob', ('eggs', '.c')),
            ])


    def test_alt_walk_func(self):
        dirnames = self.set_files(
            ('spam', ['file1.c', 'file2.c']),
            ('eggs', ['ham/file3.c']),
            )
        def get_files(root):
            return None

        files = list(iter_files(dirnames, '.c',
                                get_files=get_files,
                                _walk=self._walk,
                                _glob=self._glob))

        self.assertEqual(files, [
            fixpath('spam/file1.c'),
            fixpath('spam/file2.c'),
            fixpath('eggs/ham/file3.c'),
            ])
        self.assertEqual(self.calls, [
            ('_walk', ('spam', '.c', get_files)),
            ('_walk', ('eggs', '.c', get_files)),
            ])






#    def test_no_dirnames(self):
#        dirnames = []
#        filter_by_name = None
#
#        files = list(iter_files(dirnames, filter_by_name,
#                                _walk=self._walk))
#
#        self.assertEqual(files, [])
#        self.assertEqual(self.calls, [])
#
#    def test_no_filter(self):
#        self._return_walk = [
#                [('spam', (), ('file1', 'file2.c', 'file3.h', 'file4.o')),
#                 ],
#                ]
#        dirnames = [
#                'spam',
#                ]
#        filter_by_name = None
#
#        files = list(iter_files(dirnames, filter_by_name,
#                                _walk=self._walk))
#
#        self.assertEqual(files, [
#            fixpath('spam/file1'),
#            fixpath('spam/file2.c'),
#            fixpath('spam/file3.h'),
#            fixpath('spam/file4.o'),
#            ])
#        self.assertEqual(self.calls, [
#            ('_walk', ('spam',)),
#            ])
#
#    def test_no_files(self):
#        self._return_walk = [
#                [('spam', (), ()),
#                 ],
#                [(fixpath('eggs/ham'), (), ()),
#                 ],
#                ]
#        dirnames = [
#                'spam',
#                fixpath('eggs/ham'),
#                ]
#        filter_by_name = None
#
#        files = list(iter_files(dirnames, filter_by_name,
#                                _walk=self._walk))
#
#        self.assertEqual(files, [])
#        self.assertEqual(self.calls, [
#            ('_walk', ('spam',)),
#            ('_walk', (fixpath('eggs/ham'),)),
#            ])
#
#    def test_tree(self):
#        self._return_walk = [
#                [('spam', ('sub1', 'sub2', 'sub3'), ('file1',)),
#                 (fixpath('spam/sub1'), ('sub1sub1',), ('file2', 'file3')),
#                 (fixpath('spam/sub1/sub1sub1'), (), ('file4',)),
#                 (fixpath('spam/sub2'), (), ()),
#                 (fixpath('spam/sub3'), (), ('file5',)),
#                 ],
#                [(fixpath('eggs/ham'), (), ('file6',)),
#                 ],
#                ]
#        dirnames = [
#                'spam',
#                fixpath('eggs/ham'),
#                ]
#        filter_by_name = None
#
#        files = list(iter_files(dirnames, filter_by_name,
#                                _walk=self._walk))
#
#        self.assertEqual(files, [
#            fixpath('spam/file1'),
#            fixpath('spam/sub1/file2'),
#            fixpath('spam/sub1/file3'),
#            fixpath('spam/sub1/sub1sub1/file4'),
#            fixpath('spam/sub3/file5'),
#            fixpath('eggs/ham/file6'),
#            ])
#        self.assertEqual(self.calls, [
#            ('_walk', ('spam',)),
#            ('_walk', (fixpath('eggs/ham'),)),
#            ])
#
#    def test_filter_suffixes(self):
#        self._return_walk = [
#                [('spam', (), ('file1', 'file2.c', 'file3.h', 'file4.o')),
#                 ],
#                ]
#        dirnames = [
#                'spam',
#                ]
#        filter_by_name = ('.c', '.h')
#
#        files = list(iter_files(dirnames, filter_by_name,
#                                _walk=self._walk))
#
#        self.assertEqual(files, [
#            fixpath('spam/file2.c'),
#            fixpath('spam/file3.h'),
#            ])
#        self.assertEqual(self.calls, [
#            ('_walk', ('spam',)),
#            ])
#
#    def test_some_filtered(self):
#        self._return_walk = [
#                [('spam', (), ('file1', 'file2', 'file3', 'file4')),
#                 ],
#                ]
#        dirnames = [
#                'spam',
#                ]
#        def filter_by_name(filename, results=[False, True, False, True]):
#            self.calls.append(('filter_by_name', (filename,)))
#            return results.pop(0)
#
#        files = list(iter_files(dirnames, filter_by_name,
#                                _walk=self._walk))
#
#        self.assertEqual(files, [
#            fixpath('spam/file2'),
#            fixpath('spam/file4'),
#            ])
#        self.assertEqual(self.calls, [
#            ('_walk', ('spam',)),
#            ('filter_by_name', ('file1',)),
#            ('filter_by_name', ('file2',)),
#            ('filter_by_name', ('file3',)),
#            ('filter_by_name', ('file4',)),
#            ])
#
#    def test_none_filtered(self):
#        self._return_walk = [
#                [('spam', (), ('file1', 'file2', 'file3', 'file4')),
#                 ],
#                ]
#        dirnames = [
#                'spam',
#                ]
#        def filter_by_name(filename, results=[True, True, True, True]):
#            self.calls.append(('filter_by_name', (filename,)))
#            return results.pop(0)
#
#        files = list(iter_files(dirnames, filter_by_name,
#                                _walk=self._walk))
#
#        self.assertEqual(files, [
#            fixpath('spam/file1'),
#            fixpath('spam/file2'),
#            fixpath('spam/file3'),
#            fixpath('spam/file4'),
#            ])
#        self.assertEqual(self.calls, [
#            ('_walk', ('spam',)),
#            ('filter_by_name', ('file1',)),
#            ('filter_by_name', ('file2',)),
#            ('filter_by_name', ('file3',)),
#            ('filter_by_name', ('file4',)),
#            ])
#
#    def test_all_filtered(self):
#        self._return_walk = [
#                [('spam', (), ('file1', 'file2', 'file3', 'file4')),
#                 ],
#                ]
#        dirnames = [
#                'spam',
#                ]
#        def filter_by_name(filename, results=[False, False, False, False]):
#            self.calls.append(('filter_by_name', (filename,)))
#            return results.pop(0)
#
#        files = list(iter_files(dirnames, filter_by_name,
#                                _walk=self._walk))
#
#        self.assertEqual(files, [])
#        self.assertEqual(self.calls, [
#            ('_walk', ('spam',)),
#            ('filter_by_name', ('file1',)),
#            ('filter_by_name', ('file2',)),
#            ('filter_by_name', ('file3',)),
#            ('filter_by_name', ('file4',)),
#            ])



#!/usr/bin/env python3.8
"""Find the maximum amount of nesting for an expression that can be parsed
without causing a parse error.

Starting at the INITIAL_NESTING_DEPTH, an expression containing n parenthesis
around a 0 is generated then tested with both the C and Python parsers. We
continue incrementing the number of parenthesis by 10 until both parsers have
failed. As soon as a single parser fails, we stop testing that parser.

The grammar file, initial nesting size, and amount by which the nested size is
incremented on each success can be controlled by changing the GRAMMAR_FILE,
INITIAL_NESTING_DEPTH, or NESTED_INCR_AMT variables.

Usage: python -m scripts.find_max_nesting
"""
import sys

from _peg_parser import parse_string

GRAMMAR_FILE = "data/python.gram"
INITIAL_NESTING_DEPTH = 10
NESTED_INCR_AMT = 10


FAIL = "\033[91m"
ENDC = "\033[0m"


def check_nested_expr(nesting_depth: int) -> bool:
    expr = f"{'(' * nesting_depth}0{')' * nesting_depth}"

    try:
        parse_string(expr)
        print(f"Nesting depth of {nesting_depth} is successful")
        return True
    except Exception as err:
        print(f"{FAIL}(Failed with nesting depth of {nesting_depth}{ENDC}")
        print(f"{FAIL}\t{err}{ENDC}")
        return False


def main() -> None:
    print(f"Testing {GRAMMAR_FILE} starting at nesting depth of {INITIAL_NESTING_DEPTH}...")

    nesting_depth = INITIAL_NESTING_DEPTH
    succeeded = True
    while succeeded:
        expr = f"{'(' * nesting_depth}0{')' * nesting_depth}"
        if succeeded:
            succeeded = check_nested_expr(nesting_depth)
        nesting_depth += NESTED_INCR_AMT

    sys.exit(1)


if __name__ == "__main__":
    main()



import types
import unittest


def global_function():
    def inner_function():
        class LocalClass:
            pass
        global inner_global_function
        def inner_global_function():
            def inner_function2():
                pass
            return inner_function2
        return LocalClass
    return lambda: inner_function


class FuncAttrsTest(unittest.TestCase):
    def setUp(self):
        class F:
            def a(self):
                pass
        def b():
            return 3
        self.fi = F()
        self.F = F
        self.b = b

    def cannot_set_attr(self, obj, name, value, exceptions):
        try:
            setattr(obj, name, value)
        except exceptions:
            pass
        else:
            self.fail("shouldn't be able to set %s to %r" % (name, value))
        try:
            delattr(obj, name)
        except exceptions:
            pass
        else:
            self.fail("shouldn't be able to del %s" % name)


class FunctionPropertiesTest(FuncAttrsTest):
    # Include the external setUp method that is common to all tests
    def test_module(self):
        self.assertEqual(self.b.__module__, __name__)

    def test_dir_includes_correct_attrs(self):
        self.b.known_attr = 7
        self.assertIn('known_attr', dir(self.b),
            "set attributes not in dir listing of method")
        # Test on underlying function object of method
        self.F.a.known_attr = 7
        self.assertIn('known_attr', dir(self.fi.a), "set attribute on function "
                     "implementations, should show up in next dir")

    def test_duplicate_function_equality(self):
        # Body of `duplicate' is the exact same as self.b
        def duplicate():
            'my docstring'
            return 3
        self.assertNotEqual(self.b, duplicate)

    def test_copying___code__(self):
        def test(): pass
        self.assertEqual(test(), None)
        test.__code__ = self.b.__code__
        self.assertEqual(test(), 3) # self.b always returns 3, arbitrarily

    def test___globals__(self):
        self.assertIs(self.b.__globals__, globals())
        self.cannot_set_attr(self.b, '__globals__', 2,
                             (AttributeError, TypeError))

    def test___closure__(self):
        a = 12
        def f(): print(a)
        c = f.__closure__
        self.assertIsInstance(c, tuple)
        self.assertEqual(len(c), 1)
        # don't have a type object handy
        self.assertEqual(c[0].__class__.__name__, "cell")
        self.cannot_set_attr(f, "__closure__", c, AttributeError)

    def test_cell_new(self):
        cell_obj = types.CellType(1)
        self.assertEqual(cell_obj.cell_contents, 1)

        cell_obj = types.CellType()
        msg = "shouldn't be able to read an empty cell"
        with self.assertRaises(ValueError, msg=msg):
            cell_obj.cell_contents

    def test_empty_cell(self):
        def f(): print(a)
        try:
            f.__closure__[0].cell_contents
        except ValueError:
            pass
        else:
            self.fail("shouldn't be able to read an empty cell")
        a = 12

    def test_set_cell(self):
        a = 12
        def f(): return a
        c = f.__closure__
        c[0].cell_contents = 9
        self.assertEqual(c[0].cell_contents, 9)
        self.assertEqual(f(), 9)
        self.assertEqual(a, 9)
        del c[0].cell_contents
        try:
            c[0].cell_contents
        except ValueError:
            pass
        else:
            self.fail("shouldn't be able to read an empty cell")
        with self.assertRaises(NameError):
            f()
        with self.assertRaises(UnboundLocalError):
            print(a)

    def test___name__(self):
        self.assertEqual(self.b.__name__, 'b')
        self.b.__name__ = 'c'
        self.assertEqual(self.b.__name__, 'c')
        self.b.__name__ = 'd'
        self.assertEqual(self.b.__name__, 'd')
        # __name__ and __name__ must be a string
        self.cannot_set_attr(self.b, '__name__', 7, TypeError)
        # __name__ must be available when in restricted mode. Exec will raise
        # AttributeError if __name__ is not available on f.
        s = """def f(): pass\nf.__name__"""
        exec(s, {'__builtins__': {}})
        # Test on methods, too
        self.assertEqual(self.fi.a.__name__, 'a')
        self.cannot_set_attr(self.fi.a, "__name__", 'a', AttributeError)

    def test___qualname__(self):
        # PEP 3155
        self.assertEqual(self.b.__qualname__, 'FuncAttrsTest.setUp.<locals>.b')
        self.assertEqual(FuncAttrsTest.setUp.__qualname__, 'FuncAttrsTest.setUp')
        self.assertEqual(global_function.__qualname__, 'global_function')
        self.assertEqual(global_function().__qualname__,
                         'global_function.<locals>.<lambda>')
        self.assertEqual(global_function()().__qualname__,
                         'global_function.<locals>.inner_function')
        self.assertEqual(global_function()()().__qualname__,
                         'global_function.<locals>.inner_function.<locals>.LocalClass')
        self.assertEqual(inner_global_function.__qualname__, 'inner_global_function')
        self.assertEqual(inner_global_function().__qualname__, 'inner_global_function.<locals>.inner_function2')
        self.b.__qualname__ = 'c'
        self.assertEqual(self.b.__qualname__, 'c')
        self.b.__qualname__ = 'd'
        self.assertEqual(self.b.__qualname__, 'd')
        # __qualname__ must be a string
        self.cannot_set_attr(self.b, '__qualname__', 7, TypeError)

    def test___code__(self):
        num_one, num_two = 7, 8
        def a(): pass
        def b(): return 12
        def c(): return num_one
        def d(): return num_two
        def e(): return num_one, num_two
        for func in [a, b, c, d, e]:
            self.assertEqual(type(func.__code__), types.CodeType)
        self.assertEqual(c(), 7)
        self.assertEqual(d(), 8)
        d.__code__ = c.__code__
        self.assertEqual(c.__code__, d.__code__)
        self.assertEqual(c(), 7)
        # self.assertEqual(d(), 7)
        try:
            b.__code__ = c.__code__
        except ValueError:
            pass
        else:
            self.fail("__code__ with different numbers of free vars should "
                      "not be possible")
        try:
            e.__code__ = d.__code__
        except ValueError:
            pass
        else:
            self.fail("__code__ with different numbers of free vars should "
                      "not be possible")

    def test_blank_func_defaults(self):
        self.assertEqual(self.b.__defaults__, None)
        del self.b.__defaults__
        self.assertEqual(self.b.__defaults__, None)

    def test_func_default_args(self):
        def first_func(a, b):
            return a+b
        def second_func(a=1, b=2):
            return a+b
        self.assertEqual(first_func.__defaults__, None)
        self.assertEqual(second_func.__defaults__, (1, 2))
        first_func.__defaults__ = (1, 2)
        self.assertEqual(first_func.__defaults__, (1, 2))
        self.assertEqual(first_func(), 3)
        self.assertEqual(first_func(3), 5)
        self.assertEqual(first_func(3, 5), 8)
        del second_func.__defaults__
        self.assertEqual(second_func.__defaults__, None)
        try:
            second_func()
        except TypeError:
            pass
        else:
            self.fail("__defaults__ does not update; deleting it does not "
                      "remove requirement")


class InstancemethodAttrTest(FuncAttrsTest):

    def test___class__(self):
        self.assertEqual(self.fi.a.__self__.__class__, self.F)
        self.cannot_set_attr(self.fi.a, "__class__", self.F, TypeError)

    def test___func__(self):
        self.assertEqual(self.fi.a.__func__, self.F.a)
        self.cannot_set_attr(self.fi.a, "__func__", self.F.a, AttributeError)

    def test___self__(self):
        self.assertEqual(self.fi.a.__self__, self.fi)
        self.cannot_set_attr(self.fi.a, "__self__", self.fi, AttributeError)

    def test___func___non_method(self):
        # Behavior should be the same when a method is added via an attr
        # assignment
        self.fi.id = types.MethodType(id, self.fi)
        self.assertEqual(self.fi.id(), id(self.fi))
        # Test usage
        try:
            self.fi.id.unknown_attr
        except AttributeError:
            pass
        else:
            self.fail("using unknown attributes should raise AttributeError")
        # Test assignment and deletion
        self.cannot_set_attr(self.fi.id, 'unknown_attr', 2, AttributeError)


class ArbitraryFunctionAttrTest(FuncAttrsTest):
    def test_set_attr(self):
        self.b.known_attr = 7
        self.assertEqual(self.b.known_attr, 7)
        try:
            self.fi.a.known_attr = 7
        except AttributeError:
            pass
        else:
            self.fail("setting attributes on methods should raise error")

    def test_delete_unknown_attr(self):
        try:
            del self.b.unknown_attr
        except AttributeError:
            pass
        else:
            self.fail("deleting unknown attribute should raise TypeError")

    def test_unset_attr(self):
        for func in [self.b, self.fi.a]:
            try:
                func.non_existent_attr
            except AttributeError:
                pass
            else:
                self.fail("using unknown attributes should raise "
                          "AttributeError")


class FunctionDictsTest(FuncAttrsTest):
    def test_setting_dict_to_invalid(self):
        self.cannot_set_attr(self.b, '__dict__', None, TypeError)
        from collections import UserDict
        d = UserDict({'known_attr': 7})
        self.cannot_set_attr(self.fi.a.__func__, '__dict__', d, TypeError)

    def test_setting_dict_to_valid(self):
        d = {'known_attr': 7}
        self.b.__dict__ = d
        # Test assignment
        self.assertIs(d, self.b.__dict__)
        # ... and on all the different ways of referencing the method's func
        self.F.a.__dict__ = d
        self.assertIs(d, self.fi.a.__func__.__dict__)
        self.assertIs(d, self.fi.a.__dict__)
        # Test value
        self.assertEqual(self.b.known_attr, 7)
        self.assertEqual(self.b.__dict__['known_attr'], 7)
        # ... and again, on all the different method's names
        self.assertEqual(self.fi.a.__func__.known_attr, 7)
        self.assertEqual(self.fi.a.known_attr, 7)

    def test_delete___dict__(self):
        try:
            del self.b.__dict__
        except TypeError:
            pass
        else:
            self.fail("deleting function dictionary should raise TypeError")

    def test_unassigned_dict(self):
        self.assertEqual(self.b.__dict__, {})

    def test_func_as_dict_key(self):
        value = "Some string"
        d = {}
        d[self.b] = value
        self.assertEqual(d[self.b], value)


class FunctionDocstringTest(FuncAttrsTest):
    def test_set_docstring_attr(self):
        self.assertEqual(self.b.__doc__, None)
        docstr = "A test method that does nothing"
        self.b.__doc__ = docstr
        self.F.a.__doc__ = docstr
        self.assertEqual(self.b.__doc__, docstr)
        self.assertEqual(self.fi.a.__doc__, docstr)
        self.cannot_set_attr(self.fi.a, "__doc__", docstr, AttributeError)

    def test_delete_docstring(self):
        self.b.__doc__ = "The docstring"
        del self.b.__doc__
        self.assertEqual(self.b.__doc__, None)


def cell(value):
    """Create a cell containing the given value."""
    def f():
        print(a)
    a = value
    return f.__closure__[0]

def empty_cell(empty=True):
    """Create an empty cell."""
    def f():
        print(a)
    # the intent of the following line is simply "if False:";  it's
    # spelt this way to avoid the danger that a future optimization
    # might simply remove an "if False:" code block.
    if not empty:
        a = 1729
    return f.__closure__[0]


class CellTest(unittest.TestCase):
    def test_comparison(self):
        # These tests are here simply to exercise the comparison code;
        # their presence should not be interpreted as providing any
        # guarantees about the semantics (or even existence) of cell
        # comparisons in future versions of CPython.
        self.assertTrue(cell(2) < cell(3))
        self.assertTrue(empty_cell() < cell('saturday'))
        self.assertTrue(empty_cell() == empty_cell())
        self.assertTrue(cell(-36) == cell(-36.0))
        self.assertTrue(cell(True) > empty_cell())


class StaticMethodAttrsTest(unittest.TestCase):
    def test_func_attribute(self):
        def f():
            pass

        c = classmethod(f)
        self.assertTrue(c.__func__ is f)

        s = staticmethod(f)
        self.assertTrue(s.__func__ is f)


class BuiltinFunctionPropertiesTest(unittest.TestCase):
    # XXX Not sure where this should really go since I can't find a
    # test module specifically for builtin_function_or_method.

    def test_builtin__qualname__(self):
        import time

        # builtin function:
        self.assertEqual(len.__qualname__, 'len')
        self.assertEqual(time.time.__qualname__, 'time')

        # builtin classmethod:
        self.assertEqual(dict.fromkeys.__qualname__, 'dict.fromkeys')
        self.assertEqual(float.__getformat__.__qualname__,
                         'float.__getformat__')

        # builtin staticmethod:
        self.assertEqual(str.maketrans.__qualname__, 'str.maketrans')
        self.assertEqual(bytes.maketrans.__qualname__, 'bytes.maketrans')

        # builtin bound instance method:
        self.assertEqual([1, 2, 3].append.__qualname__, 'list.append')
        self.assertEqual({'foo': 'bar'}.pop.__qualname__, 'dict.pop')


if __name__ == "__main__":
    unittest.main()



"""Python parser generator


This parser generator transforms a Python grammar file into parsing tables
that can be consumed by Python's LL(1) parser written in C.

Concepts
--------

* An LL(1) parser (Left-to-right, Leftmost derivation, 1 token-lookahead) is a
  top-down parser for a subset of context-free languages. It parses the input
  from Left to right, performing Leftmost derivation of the sentence, and can
  only use 1 token of lookahead when parsing a sentence.

* A parsing table is a collection of data that a generic implementation of the
  LL(1) parser consumes to know how to parse a given context-free grammar. In
  this case the collection of data involves Deterministic Finite Automatons,
  calculated first sets, keywords and transition labels.

* A grammar is defined by production rules (or just 'productions') that specify
  which symbols may replace which other symbols; these rules may be used to
  generate strings, or to parse them. Each such rule has a head, or left-hand
  side, which consists of the string that may be replaced, and a body, or
  right-hand side, which consists of a string that may replace it. In the
  Python grammar, rules are written in the form

  rule_name: rule_description;

  meaning the rule 'a: b' specifies that a can be replaced by b. A context-free
  grammar is a grammar in which the left-hand side of each production rule
  consists of only a single nonterminal symbol. Context-free grammars can
  always be recognized by a Non-Deterministic Automatons.

* Terminal symbols are literal symbols which may appear in the outputs of the
  production rules of the grammar and which cannot be changed using the rules
  of the grammar. Applying the rules recursively to a source string of symbols
  will usually terminate in a final output string consisting only of terminal
  symbols.

* Nonterminal symbols are those symbols which can be replaced. The grammar
  includes a start symbol a designated member of the set of nonterminals from
  which all the strings in the language may be derived by successive
  applications of the production rules.

* The language defined by the grammar is defined as the set of terminal strings
  that can be derived using the production rules.

* The first sets of a rule (FIRST(rule)) are defined to be the set of terminals
  that can appear in the first position of any string derived from the rule.
  This is useful for LL(1) parsers as the parser is only allowed to look at the
  next token in the input to know which rule needs to parse. For example, given
  this grammar:

  start: '(' A | B ')'
  A: 'a' '<'
  B: 'b' '<'

  and the input '(b<)' the parser can only look at 'b' to know if it needs
  to parse A o B. Because FIRST(A) = {'a'} and FIRST(B) = {'b'} it knows
  that needs to continue parsing rule B because only that rule can start
  with 'b'.

Description
-----------

The input for the parser generator is a grammar in extended BNF form (using *
for repetition, + for at-least-once repetition, [] for optional parts, | for
alternatives and () for grouping).

Each rule in the grammar file is considered as a regular expression in its
own right. It is turned into a Non-deterministic Finite Automaton (NFA),
which is then turned into a Deterministic Finite Automaton (DFA), which is
then optimized to reduce the number of states. See [Aho&Ullman 77] chapter 3,
or similar compiler books (this technique is more often used for lexical
analyzers).

The DFA's are used by the parser as parsing tables in a special way that's
probably unique. Before they are usable, the FIRST sets of all non-terminals
are computed so the LL(1) parser consuming the parsing tables can distinguish
between different transitions.
Reference
---------

[Aho&Ullman 77]
    Aho&Ullman, Principles of Compiler Design, Addison-Wesley 1977
    (first edition)
"""

from ast import literal_eval
import collections

from . import grammar, token
from .automata import DFA
from .metaparser import GrammarParser

import enum


class LabelType(enum.Enum):
    NONTERMINAL = 0
    NAMED_TOKEN = 1
    KEYWORD = 2
    OPERATOR = 3
    NONE = 4


class Label(str):
    def __init__(self, value):
        self.type = self._get_type()

    def _get_type(self):
        if self[0].isalpha():
            if self.upper() == self:
                # NAMED tokens (ASYNC, NAME...) are all uppercase by convention
                return LabelType.NAMED_TOKEN
            else:
                # If is not uppercase it must be a non terminal.
                return LabelType.NONTERMINAL
        else:
            # Keywords and operators are wrapped in quotes
            assert self[0] == self[-1] in ('"', "'"), self
            value = literal_eval(self)
            if value[0].isalpha():
                return LabelType.KEYWORD
            else:
                return LabelType.OPERATOR

    def __repr__(self):
        return "{}({})".format(self.type, super().__repr__())


class ParserGenerator(object):
    def __init__(self, grammar_file, token_file, verbose=False, graph_file=None):
        with open(grammar_file) as f:
            self.grammar = f.read()
        with open(token_file) as tok_file:
            token_lines = tok_file.readlines()
        self.tokens = dict(token.generate_tokens(token_lines))
        self.opmap = dict(token.generate_opmap(token_lines))
        # Manually add <> so it does not collide with !=
        self.opmap["<>"] = "NOTEQUAL"
        self.verbose = verbose
        self.filename = grammar_file
        self.graph_file = graph_file
        self.dfas, self.startsymbol = self.create_dfas()
        self.first = {}  # map from symbol name to set of tokens
        self.calculate_first_sets()

    def create_dfas(self):
        rule_to_dfas = collections.OrderedDict()
        start_nonterminal = None
        for nfa in GrammarParser(self.grammar).parse():
            if self.verbose:
                print("Dump of NFA for", nfa.name)
                nfa.dump()
            if self.graph_file is not None:
                nfa.dump_graph(self.graph_file.write)
            dfa = DFA.from_nfa(nfa)
            if self.verbose:
                print("Dump of DFA for", dfa.name)
                dfa.dump()
            dfa.simplify()
            if self.graph_file is not None:
                dfa.dump_graph(self.graph_file.write)
            rule_to_dfas[dfa.name] = dfa

            if start_nonterminal is None:
                start_nonterminal = dfa.name

        return rule_to_dfas, start_nonterminal

    def make_grammar(self):
        c = grammar.Grammar()
        c.all_labels = set()
        names = list(self.dfas.keys())
        names.remove(self.startsymbol)
        names.insert(0, self.startsymbol)
        for name in names:
            i = 256 + len(c.symbol2number)
            c.symbol2number[Label(name)] = i
            c.number2symbol[i] = Label(name)
            c.all_labels.add(name)
        for name in names:
            self.make_label(c, name)
            dfa = self.dfas[name]
            states = []
            for state in dfa:
                arcs = []
                for label, next in sorted(state.arcs.items()):
                    c.all_labels.add(label)
                    arcs.append((self.make_label(c, label), dfa.states.index(next)))
                if state.is_final:
                    arcs.append((0, dfa.states.index(state)))
                states.append(arcs)
            c.states.append(states)
            c.dfas[c.symbol2number[name]] = (states, self.make_first_sets(c, name))
        c.start = c.symbol2number[self.startsymbol]

        if self.verbose:
            print("")
            print("Grammar summary")
            print("===============")

            print("- {n_labels} labels".format(n_labels=len(c.labels)))
            print("- {n_dfas} dfas".format(n_dfas=len(c.dfas)))
            print("- {n_tokens} tokens".format(n_tokens=len(c.tokens)))
            print("- {n_keywords} keywords".format(n_keywords=len(c.keywords)))
            print(
                "- Start symbol: {start_symbol}".format(
                    start_symbol=c.number2symbol[c.start]
                )
            )
        return c

    def make_first_sets(self, c, name):
        rawfirst = self.first[name]
        first = set()
        for label in sorted(rawfirst):
            ilabel = self.make_label(c, label)
            ##assert ilabel not in first # XXX failed on <> ... !=
            first.add(ilabel)
        return first

    def make_label(self, c, label):
        label = Label(label)
        ilabel = len(c.labels)

        if label.type == LabelType.NONTERMINAL:
            if label in c.symbol2label:
                return c.symbol2label[label]
            else:
                c.labels.append((c.symbol2number[label], None))
                c.symbol2label[label] = ilabel
                return ilabel
        elif label.type == LabelType.NAMED_TOKEN:
            # A named token (NAME, NUMBER, STRING)
            itoken = self.tokens.get(label, None)
            assert isinstance(itoken, int), label
            assert itoken in self.tokens.values(), label
            if itoken in c.tokens:
                return c.tokens[itoken]
            else:
                c.labels.append((itoken, None))
                c.tokens[itoken] = ilabel
                return ilabel
        elif label.type == LabelType.KEYWORD:
            # A keyword
            value = literal_eval(label)
            if value in c.keywords:
                return c.keywords[value]
            else:
                c.labels.append((self.tokens["NAME"], value))
                c.keywords[value] = ilabel
                return ilabel
        elif label.type == LabelType.OPERATOR:
            # An operator (any non-numeric token)
            value = literal_eval(label)
            tok_name = self.opmap[value]  # Fails if unknown token
            itoken = self.tokens[tok_name]
            if itoken in c.tokens:
                return c.tokens[itoken]
            else:
                c.labels.append((itoken, None))
                c.tokens[itoken] = ilabel
                return ilabel
        else:
            raise ValueError("Cannot categorize label {}".format(label))

    def calculate_first_sets(self):
        names = list(self.dfas.keys())
        for name in names:
            if name not in self.first:
                self.calculate_first_sets_for_rule(name)

            if self.verbose:
                print("First set for {dfa_name}".format(dfa_name=name))
                for item in self.first[name]:
                    print("    - {terminal}".format(terminal=item))

    def calculate_first_sets_for_rule(self, name):
        dfa = self.dfas[name]
        self.first[name] = None  # dummy to detect left recursion
        state = dfa.states[0]
        totalset = set()
        overlapcheck = {}
        for label, next in state.arcs.items():
            if label in self.dfas:
                if label in self.first:
                    fset = self.first[label]
                    if fset is None:
                        raise ValueError("recursion for rule %r" % name)
                else:
                    self.calculate_first_sets_for_rule(label)
                    fset = self.first[label]
                totalset.update(fset)
                overlapcheck[label] = fset
            else:
                totalset.add(label)
                overlapcheck[label] = {label}
        inverse = {}
        for label, itsfirst in overlapcheck.items():
            for symbol in itsfirst:
                if symbol in inverse:
                    raise ValueError(
                        "rule %s is ambiguous; %s is in the"
                        " first sets of %s as well as %s"
                        % (name, symbol, label, inverse[symbol])
                    )
                inverse[symbol] = label
        self.first[name] = totalset



from collections import namedtuple
import shlex
import os
import re

from ..common import util, info


CONTINUATION = '\\' + os.linesep

IDENTIFIER = r'(?:\w*[a-zA-Z]\w*)'
IDENTIFIER_RE = re.compile('^' + IDENTIFIER + '$')


def _coerce_str(value):
    if not value:
        return ''
    return str(value).strip()


#############################
# directives

DIRECTIVE_START = r'''
    (?:
      ^ \s*
      [#] \s*
      )'''
DIRECTIVE_TEXT = r'''
    (?:
      (?: \s+ ( .*\S ) )?
      \s* $
      )'''
DIRECTIVE = rf'''
    (?:
      {DIRECTIVE_START}
      (
        include |
        error | warning |
        pragma |
        define | undef |
        if | ifdef | ifndef | elseif | else | endif |
        __FILE__ | __LINE__ | __DATE __ | __TIME__ | __TIMESTAMP__
        )
      {DIRECTIVE_TEXT}
      )'''
#       (?:
#        [^\\\n] |
#        \\ [^\n] |
#        \\ \n
#        )+
#      ) \n
#     )'''
DIRECTIVE_RE = re.compile(DIRECTIVE, re.VERBOSE)

DEFINE = rf'''
    (?:
      {DIRECTIVE_START} define \s+
      (?:
        ( \w*[a-zA-Z]\w* )
        (?: \s* [(] ([^)]*) [)] )?
        )
      {DIRECTIVE_TEXT}
      )'''
DEFINE_RE = re.compile(DEFINE, re.VERBOSE)


def parse_directive(line):
    """Return the appropriate directive for the given line."""
    line = line.strip()
    if line.startswith('#'):
        line = line[1:].lstrip()
        line = '#' + line
    directive = line
    #directive = '#' + line
    while '  ' in directive:
        directive = directive.replace('  ', ' ')
    return _parse_directive(directive)


def _parse_directive(line):
    m = DEFINE_RE.match(line)
    if m:
        name, args, text = m.groups()
        if args:
            args = [a.strip() for a in args.split(',')]
            return Macro(name, args, text)
        else:
            return Constant(name, text)

    m = DIRECTIVE_RE.match(line)
    if not m:
        raise ValueError(f'unsupported directive {line!r}')
    kind, text = m.groups()
    if not text:
        if kind not in ('else', 'endif'):
            raise ValueError(f'missing text in directive {line!r}')
    elif kind in ('else', 'endif', 'define'):
        raise ValueError(f'unexpected text in directive {line!r}')
    if kind == 'include':
        directive = Include(text)
    elif kind in IfDirective.KINDS:
        directive = IfDirective(kind, text)
    else:
        directive = OtherDirective(kind, text)
    directive.validate()
    return directive


class PreprocessorDirective(util._NTBase):
    """The base class for directives."""

    __slots__ = ()

    KINDS = frozenset([
            'include',
            'pragma',
            'error', 'warning',
            'define', 'undef',
            'if', 'ifdef', 'ifndef', 'elseif', 'else', 'endif',
            '__FILE__', '__DATE__', '__LINE__', '__TIME__', '__TIMESTAMP__',
            ])

    @property
    def text(self):
        return ' '.join(v for v in self[1:] if v and v.strip()) or None

    def validate(self):
        """Fail if the object is invalid (i.e. init with bad data)."""
        super().validate()

        if not self.kind:
            raise TypeError('missing kind')
        elif self.kind not in self.KINDS:
            raise ValueError

        # text can be anything, including None.


class Constant(PreprocessorDirective,
               namedtuple('Constant', 'kind name value')):
    """A single "constant" directive ("define")."""

    __slots__ = ()

    def __new__(cls, name, value=None):
        self = super().__new__(
                cls,
                'define',
                name=_coerce_str(name) or None,
                value=_coerce_str(value) or None,
                )
        return self

    def validate(self):
        """Fail if the object is invalid (i.e. init with bad data)."""
        super().validate()

        if not self.name:
            raise TypeError('missing name')
        elif not IDENTIFIER_RE.match(self.name):
            raise ValueError(f'name must be identifier, got {self.name!r}')

        # value can be anything, including None


class Macro(PreprocessorDirective,
            namedtuple('Macro', 'kind name args body')):
    """A single "macro" directive ("define")."""

    __slots__ = ()

    def __new__(cls, name, args, body=None):
        # "args" must be a string or an iterable of strings (or "empty").
        if isinstance(args, str):
            args = [v.strip() for v in args.split(',')]
        if args:
            args = tuple(_coerce_str(a) or None for a in args)
        self = super().__new__(
                cls,
                kind='define',
                name=_coerce_str(name) or None,
                args=args if args else (),
                body=_coerce_str(body) or None,
                )
        return self

    @property
    def text(self):
        if self.body:
            return f'{self.name}({", ".join(self.args)}) {self.body}'
        else:
            return f'{self.name}({", ".join(self.args)})'

    def validate(self):
        """Fail if the object is invalid (i.e. init with bad data)."""
        super().validate()

        if not self.name:
            raise TypeError('missing name')
        elif not IDENTIFIER_RE.match(self.name):
            raise ValueError(f'name must be identifier, got {self.name!r}')

        for arg in self.args:
            if not arg:
                raise ValueError(f'missing arg in {self.args}')
            elif not IDENTIFIER_RE.match(arg):
                raise ValueError(f'arg must be identifier, got {arg!r}')

        # body can be anything, including None


class IfDirective(PreprocessorDirective,
                  namedtuple('IfDirective', 'kind condition')):
    """A single conditional directive (e.g. "if", "ifdef").

    This only includes directives that actually provide conditions.  The
    related directives "else" and "endif" are covered by OtherDirective
    instead.
    """

    __slots__ = ()

    KINDS = frozenset([
            'if',
            'ifdef',
            'ifndef',
            'elseif',
            ])

    @classmethod
    def _condition_from_raw(cls, raw, kind):
        #return Condition.from_raw(raw, _kind=kind)
        condition = _coerce_str(raw)
        if not condition:
            return None

        if kind == 'ifdef':
            condition = f'defined({condition})'
        elif kind == 'ifndef':
            condition = f'! defined({condition})'

        return condition

    def __new__(cls, kind, condition):
        kind = _coerce_str(kind)
        self = super().__new__(
                cls,
                kind=kind or None,
                condition=cls._condition_from_raw(condition, kind),
                )
        return self

    @property
    def text(self):
        if self.kind == 'ifdef':
            return self.condition[8:-1]  # strip "defined("
        elif self.kind == 'ifndef':
            return self.condition[10:-1]  # strip "! defined("
        else:
            return self.condition
        #return str(self.condition)

    def validate(self):
        """Fail if the object is invalid (i.e. init with bad data)."""
        super().validate()

        if not self.condition:
            raise TypeError('missing condition')
        #else:
        #    for cond in self.condition:
        #        if not cond:
        #            raise ValueError(f'missing condition in {self.condition}')
        #        cond.validate()
        #    if self.kind in ('ifdef', 'ifndef'):
        #        if len(self.condition) != 1:
        #            raise ValueError('too many condition')
        #        if self.kind == 'ifdef':
        #            if not self.condition[0].startswith('defined '):
        #                raise ValueError('bad condition')
        #        else:
        #            if not self.condition[0].startswith('! defined '):
        #                raise ValueError('bad condition')


class Include(PreprocessorDirective,
              namedtuple('Include', 'kind file')):
    """A single "include" directive.

    Supported "file" values are either follow the bracket style
    (<stdio>) or double quotes ("spam.h").
    """

    __slots__ = ()

    def __new__(cls, file):
        self = super().__new__(
                cls,
                kind='include',
                file=_coerce_str(file) or None,
                )
        return self

    def validate(self):
        """Fail if the object is invalid (i.e. init with bad data)."""
        super().validate()

        if not self.file:
            raise TypeError('missing file')


class OtherDirective(PreprocessorDirective,
                     namedtuple('OtherDirective', 'kind text')):
    """A single directive not covered by another class.

    This includes the "else", "endif", and "undef" directives, which are
    otherwise inherently related to the directives covered by the
    Constant, Macro, and IfCondition classes.

    Note that all directives must have a text value, except for "else"
    and "endif" (which must have no text).
    """

    __slots__ = ()

    KINDS = PreprocessorDirective.KINDS - {'include', 'define'} - IfDirective.KINDS

    def __new__(cls, kind, text):
        self = super().__new__(
                cls,
                kind=_coerce_str(kind) or None,
                text=_coerce_str(text) or None,
                )
        return self

    def validate(self):
        """Fail if the object is invalid (i.e. init with bad data)."""
        super().validate()

        if self.text:
            if self.kind in ('else', 'endif'):
                raise ValueError('unexpected text in directive')
        elif self.kind not in ('else', 'endif'):
            raise TypeError('missing text')


#############################
# iterating lines

def _recompute_conditions(directive, ifstack):
    if directive.kind in ('if', 'ifdef', 'ifndef'):
        ifstack.append(
                ([], directive.condition))
    elif directive.kind == 'elseif':
        if ifstack:
            negated, active = ifstack.pop()
            if active:
                negated.append(active)
        else:
            negated = []
        ifstack.append(
                (negated, directive.condition))
    elif directive.kind == 'else':
        if ifstack:
            negated, active = ifstack.pop()
            if active:
                negated.append(active)
            ifstack.append(
                    (negated, None))
    elif directive.kind == 'endif':
        if ifstack:
            ifstack.pop()

    conditions = []
    for negated, active in ifstack:
        for condition in negated:
            conditions.append(f'! ({condition})')
        if active:
            conditions.append(active)
    return tuple(conditions)


def _iter_clean_lines(lines):
    lines = iter(enumerate(lines, 1))
    for lno, line in lines:
        # Handle line continuations.
        while line.endswith(CONTINUATION):
            try:
                lno, _line = next(lines)
            except StopIteration:
                break
            line = line[:-len(CONTINUATION)] + ' ' + _line

        # Deal with comments.
        after = line
        line = ''
        while True:
            # Look for a comment.
            before, begin, remainder = after.partition('/*')
            if '//' in before:
                before, _, _ = before.partition('//')
                line += before + ' '  # per the C99 spec
                break
            line += before
            if not begin:
                break
            line += ' '  # per the C99 spec

            # Go until we find the end of the comment.
            _, end, after = remainder.partition('*/')
            while not end:
                try:
                    lno, remainder = next(lines)
                except StopIteration:
                    raise Exception('unterminated comment')
                _, end, after = remainder.partition('*/')

        yield lno, line


def iter_lines(lines, *,
                   _iter_clean_lines=_iter_clean_lines,
                   _parse_directive=_parse_directive,
                   _recompute_conditions=_recompute_conditions,
                   ):
    """Yield (lno, line, directive, active conditions) for each given line.

    This is effectively a subset of the operations taking place in
    translation phases 2-4 from the C99 spec (ISO/IEC 9899:TC2); see
    section 5.1.1.2.  Line continuations are removed and comments
    replaced with a single space.  (In both cases "lno" will be the last
    line involved.)  Otherwise each line is returned as-is.

    "lno" is the (1-indexed) line number for the line.

    "directive" will be a PreprocessorDirective or None, depending on
    whether or not there is a directive on the line.

    "active conditions" is the set of preprocessor conditions (e.g.
    "defined()") under which the current line of code will be included
    in compilation.  That set is derived from every conditional
    directive block (e.g. "if defined()", "ifdef", "else") containing
    that line.  That includes nested directives.  Note that the
    current line does not affect the active conditions for iteself.
    It only impacts subsequent lines.  That applies to directives
    that close blocks (e.g. "endif") just as much as conditional
    directvies.  Also note that "else" and "elseif" directives
    update the active conditions (for later lines), rather than
    adding to them.
    """
    ifstack = []
    conditions = ()
    for lno, line in _iter_clean_lines(lines):
        stripped = line.strip()
        if not stripped.startswith('#'):
            yield lno, line, None, conditions
            continue

        directive = '#' + stripped[1:].lstrip()
        while '  ' in directive:
            directive = directive.replace('  ', ' ')
        directive = _parse_directive(directive)
        yield lno, line, directive, conditions

        if directive.kind in ('else', 'endif'):
            conditions = _recompute_conditions(directive, ifstack)
        elif isinstance(directive, IfDirective):
            conditions = _recompute_conditions(directive, ifstack)


#############################
# running (platform-specific?)

def _gcc(filename, *,
         _get_argv=(lambda: _get_gcc_argv()),
         _run=util.run_cmd,
         ):
    argv = _get_argv()
    argv.extend([
            '-E', filename,
            ])
    output = _run(argv)
    return output


def _get_gcc_argv(*,
                  _open=open,
                  _run=util.run_cmd,
                  ):
    with _open('/tmp/print.mk', 'w') as tmpfile:
        tmpfile.write('print-%:\n')
        #tmpfile.write('\t@echo $* = $($*)\n')
        tmpfile.write('\t@echo $($*)\n')
    argv = ['/usr/bin/make',
            '-f', 'Makefile',
            '-f', '/tmp/print.mk',
            'print-CC',
            'print-PY_CORE_CFLAGS',
            ]
    output = _run(argv)
    gcc, cflags = output.strip().splitlines()
    argv = shlex.split(gcc.strip())
    cflags = shlex.split(cflags.strip())
    return argv + cflags


def run(filename, *,
        _gcc=_gcc,
        ):
    """Return the text of the given file after running the preprocessor."""
    return _gcc(filename)
